import sys
import random
import logging
import argparse
import time
import json

import zmq

from package.shared.util import parse_es_result, base_dir, server_cfg_file
from package.shared.logger import setup_logger
from package.shared.config import Config

import platform

osname = sys.platform
arch = platform.machine()
dirname = '{}-{}'.format(osname, arch)
fullpath = '{}/bin/{}'.format(base_dir(), dirname)
sys.path.append(fullpath)

import pylibm
sys.path.append('{}/package/esproto'.format(base_dir()))

from common_pb2 import USER_SIGN_IN, USER_OBJ_GET, USER_OBJ_CHG, OK, Val
from userapi.system_pb2 import SignInReq
from userapi.object_pb2 import ObjGetReq, ObjGetResp, Obj


class ESSendError(Exception):
    pass

class IntializationError(Exception):
    pass

class ServiceLoginError(Exception):
    pass

class EventServerConnectionError(Exception):
    pass

class ESProtoError(Exception):
    pass


class ESConnection:
    def __init__(self, args):
        config = Config.get()
        self.m_args = args

        self.ctx = pylibm.mc_ctx()
        self.conn = pylibm.mc_conn(self.ctx)
        self.op_timeout = 15000
        self.login_request_id = 0
        self.request_objects_request_id = 0
        self.logged_in = False
        self.initialized = False

        self.conn.set_msg_sent_cb(self.__on_message_sent)
        self.conn.set_conn_error_cb(self.__on_conn_error)
        self.conn.set_resp_recv_cb(self.__on_resp_recv)
        self.conn.set_conn_state_cb(self.__on_conn_state_change)
        self.conn.set_post_recv_cb(self.__on_post_recv)

        self.m_host = self.m_args['host']
        self.m_port = int(self.m_args['port'])

        # ZMQ
        self.m_ctx = zmq.Context()
        self.m_dealer = self.m_ctx.socket(zmq.DEALER)
        url = config.app.ipc
        self.m_dealer.setsockopt(zmq.IDENTITY, b'es')

        print(f'Connect to: {url}')
        self.m_dealer.connect(url)


    def __del__(self):
        try:
            self.conn.close()
        finally:
            self.conn.abort()


    def print_error(self, err_code):
        s = "Unknown error"
        if err_code == pylibm.MC_CONN_ERROR_OTHER:
            s = "Other connection error"
        elif err_code == pylibm.MC_CONN_ERROR_TIMEOUT:
            s = "Connection has timed out"
        elif err_code == pylibm.MC_CONN_ERROR_NOT_FOUND:
            s = "Host address was not found"
        elif err_code == pylibm.MC_CONN_ERROR_REFUSED:
            s = "TCP connection was refused by the host"
        elif err_code == pylibm.MC_CONN_ERROR_REJECTED:
            s = "Connection was rejected by the server"
        elif err_code == pylibm.MC_CONN_ERROR_CLOSED:
            s = "Connection was unexpectedly closed by the server"
        elif err_code == pylibm.MC_CONN_ERROR_ABORTED:
            s = "Connection was aborted by the client"
        elif err_code == pylibm.MC_CONN_ERROR_PROTOCOL:
            s = "Protocol error"
        elif err_code == pylibm.MC_CONN_ERROR_NETWORK:
            s = "Network error"
        return s


    def __open_connection(self):
        is_opened = False
        try:
            self.conn.open(self.m_host, self.m_port)
            #self.conn.wait_state(pylibm.MC_CONN_STATE_OPEN, self.op_timeout)
            is_opened = True
        except pylibm.error as err:
            logging.debug("Failed to open connection to ES server: {0}".format(str(err)))
            self.conn.abort()

        if not is_opened:
            raise EventServerConnectionError()


    def __login(self):
        try:
            self.__logged_in = False
            logging.info("Log in to ES server...")

            m, msg_type, self.login_request_id = self.__build_login_message() # store login request ID to identify it in callback
            self.conn.req_msg(m, msg_type, self.login_request_id, self.op_timeout)
            logging.info("Sing In request Sent")

        except pylibm.error as err:
            logging.debug("ParkingSensors Sync: {0}".format(str(err)))
            # raise LoginError()


    def __request_objects(self):
        try:
            logging.info("Requesting objects...")

            m, msg_type, self.request_objects_request_id = self.__build_request_objects_message() #store login request ID to identify it in callback
            self.conn.req_msg(m, msg_type, self.request_objects_request_id, self.op_timeout)
            logging.info("Request objects request Sent")

        except pylibm.error as err:
            logging.debug("ParkingSensors Sync: {0}".format(str(err)))
            # raise LoginError()


    def initialize(self):
        try:
            self.initialized = False
            self.__open_connection()
            self.initialized = True
        except EventServerConnectionError:
            logging.error("Failed to connect to Event Server. Make sure Event Server is available.")
        except ServiceLoginError:
            logging.error("Failed to Login to Event Server. Make sure you service login key is correct.")
        finally:
            if not self.initialized:
                raise IntializationError("Failed to initialize Trap Sender component.")


    def __build_login_message(self):
        logging.info("Prepare singIn request")

        r = SignInReq()
        r.method = SignInReq.PLAIN
        r.user = self.m_args['username']
        r.passwd = self.m_args['password']
        data = bytes(r.SerializeToString())
        # logging.debug(r.SerializeToString())

        m = pylibm.mc_msg(self.ctx, len(data))
        m.set_data(data)

        msg_type = USER_SIGN_IN
        msg_id = random.randrange(1000000)

        return m, msg_type, msg_id


    def __build_request_objects_message(self):
        logging.info("Prepare get objects request")

        r = ObjGetReq()
        r.flags.get_sens = True
        r.flags.get_actr = True
        r.flags.get_strm = True
        r.flags.get_obj_conf = True
        logging.info("Prepared get objects request")

        data = bytes(r.SerializeToString())
        m = pylibm.mc_msg(self.ctx, len(data))

        m.set_data(data)
        msg_type = USER_OBJ_GET
        msg_id = random.randrange(1000000)

        self.id_es_object = {}
        return m, msg_type, msg_id


    def __process_objects(self, msg, result):
        logging.info("Processing objects")
        logging.info("Request objects result:  {0}".format(result))
        if result != OK:
            logging.error("Request for user objects failed")
            return

        reply = ObjGetResp()
        reply.ParseFromString(msg.get_data())

        obj_by_id = parse_es_result(reply, Val, Obj)

        pkt = {
            'cmd': 'es-data',
            'objs': obj_by_id
        }

        data = json.dumps(pkt)
        self.m_dealer.send(data.encode())


    # Callbacks
    def __on_resp_recv(self, conn, msg, type, result, req_id):
        logging.info("Response for Request with ID = {0} received.".format(req_id))
        try:
            if req_id == self.login_request_id:
                logging.info("Logged in.")
                self.logged_in = True
                self.__request_objects()

            if req_id == self.request_objects_request_id:
                logging.info("Received user objects")
                self.__process_objects(msg, result)
        except Exception as e:
            logging.exception(e)


    def __on_post_recv(self, conn, msg, type):
        logging.info("Post message received type = {0}, bytes = {1}".format(type, msg.get_size()))

        try:
            if type == USER_OBJ_CHG:
                logging.info('Objects changed')
                self.__request_objects()
        except Exception as e:
            logging.exception(e)


    def __on_message_sent(self, conn, req_id):
        if req_id == self.login_request_id:
            logging.info("Login request has been sent to ES server.")
        else:
            logging.info("Message with ID = {0} has been sent to ES".format(req_id))


    def __on_conn_error(self, c, err_code):
        logging.error("Connection error occured: {0}".format(self.print_error(err_code)))
        time.sleep(5)


    def __on_conn_state_change(self, conn, state_to, state_from):
        if state_to == pylibm.MC_CONN_STATE_CLOSED:
            logging.info("Connection to ES server is Closed")
            time.sleep(5)
            self.initialize()
        elif state_to == pylibm.MC_CONN_STATE_OPEN:
            logging.info("Connection to ES server is Opened")
            self.__login()
        elif state_to == pylibm.MC_CONN_STATE_OPENING:
            logging.info("Opening connection to ES server...")
        elif state_to == pylibm.MC_CONN_STATE_CLOSING:
            logging.info("Closing connection to ES server...")


def run_es(**kwargs):
    Config.init(server_cfg_file())
    setup_logger('es.txt')

    try:
        time.sleep(3)
    except KeyboardInterrupt:
        return

    logging.info('ES is running')
    args = vars(parse_args())
    for k, v in kwargs.items():
        args[k] = v

    connection = ESConnection(args)
    connection.initialize()

    try:
        while 1:
            time.sleep(1)
    except KeyboardInterrupt:
        pass
    finally:
        del connection
        sys.exit(0)

    print('Shoulnd be here (ES)')

def parse_args():
    config = Config.get()
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', help='ES address', default=config.es.host)
    parser.add_argument('--port', help='ES port', default=config.es.port)
    parser.add_argument('--username', help='Username', default=config.es.username)
    parser.add_argument('--password', help='Password', default=config.es.password)
    return parser.parse_args()


if __name__ == '__main__':
    run_es()