from typing import Dict, List, Any, BinaryIO

import re
import tempfile
import shutil
import hashlib
import string
import codecs
import functools
from dataclasses import dataclass, field

import argparse
import datetime
import logging
import asyncio
from concurrent.futures import ThreadPoolExecutor

import smb
import smb.base
import smb.SMBConnection

import zmq
from zmq.asyncio import Context as ZmqContext
from zmq.asyncio import Socket as ZmqSocket

import sys

# import os
# os.environ['CONF_FILE_NAME'] = 'server.yaml'

from package.shared.logger import setup_logger
from package.shared.models import SMBFile, Stream, SessionSingleton
from package.shared.config import Config
from package.shared.constants import Constants
from package.shared.flow_control import FlowControl
from package.shared.util import server_cfg_file

from .pos_server import PosServer


@dataclass
class Samba():
    username: str
    password: str
    host: str
    port: int
    remote_name: str
    domain: str
    service: str
    client_name: str = 'aigsgClient'
    timeout: int = 30
    conn: smb.SMBConnection.SMBConnection = None

    def connect(self):
        if not self.conn is None:
            logging.warning('Connection initialized already!')

        logging.info('Going to connect samba on: {}:{}'.format(self.host, self.port))

        ret = False
        try:
            self.conn = smb.SMBConnection.SMBConnection(self.username,
                self.password,
                self.client_name,
                self.remote_name,
                self.domain,
                is_direct_tcp=True)

            ret = self.conn.connect(self.host, self.port)

            if not ret:
                logging.warning("Can't connect to remote host: {}".format(self.host))
        except Exception as e:
            logging.warning('Conn Exception: {}'.format(e))
            logging.warning('Type: {}'.format(type(e)))

        return ret


    def list_files(self, path='/', recursive=True):
        files = []

        if self.conn is None:
            logging.warning('No connection to list files!')
            return files

        try:
            resp = self.conn.listPath(self.service, path, timeout=self.timeout)
            for item in resp:
                item: smb.base.SharedFile
                if item.filename == '.' or item.filename == '..':
                    continue

                full_path = '{}/{}'.format(path if path!='/' else '', item.filename)
                if item.isDirectory:
                    if not recursive:
                        continue

                    files = files + self.list_files(path=full_path)
                    continue

                if item.file_size == 0:
                    logging.info('Skipping empty file: {}'.format(item.filename))
                    continue

                logging.debug('{}'.format(full_path))
                logging.debug('  Size   : {}'.format(item.file_size))
                logging.debug('  Created: {}'.format(item.create_time))
                logging.debug('  Access : {}'.format(item.last_access_time))
                logging.debug('  Write  : {}'.format(item.last_write_time))

                obj = {
                    'path': full_path,
                    'size': item.file_size,
                    'created': item.create_time,
                    'last_access': item.last_access_time,
                    'last_write': item.last_write_time
                }

                files.append(obj)
        except smb.base.NotConnectedError:
            logging.warning('Not connected yet')
        except Exception as e:
            logging.warning('List Exception: {}'.format(e))
            logging.warning('Type: {}'.format(type(e)))

        return files


    def retrieve_file(self, full_path, fileobj):
        if self.conn is None:
            logging.warning('No connection to list files!')
            return None

        file_attributes, filesize = self.conn.retrieveFile(self.service, full_path, fileobj)


def samba_retrieval(username, password, host, port, remote_name, domain, service, path, localpath):
    client = Samba(
        username=username,
        password=password,
        host=host,
        port=port,
        remote_name=remote_name,
        domain=domain,
        service=service)

    client.connect()

    files = client.list_files(path)
    for f in files:
        file_name = f['path'].replace('/', '_')
        full_path = '{}/{}'.format(localpath, file_name)

        with open(full_path, 'wb') as fp:
            client.retrieve_file(f['path'], fp)

        f['local_path'] = full_path

    return files


@dataclass
class App():
    executor: ThreadPoolExecutor = None
    posserver_by_name: Dict[str, PosServer] = None


    def __post_init__(self):
        self.posserver_by_name = dict()
        self.excutor = ThreadPoolExecutor(max_workers=1)


    async def process_files(self, files):
        logging.info('Retrieved {} files'.format(len(files)))
        session = SessionSingleton.instance().session
        config = Config.get()

        servers = config.servers

        for f in files:
            remote_path = f['path']
            filename = remote_path.split('/')[-1]
            local_path = f['local_path']

            md5 = ''
            with open(local_path, 'rb') as fp:
                md5_hash = hashlib.md5()
                for chunk in iter(lambda: fp.read(4096), b""):
                    md5_hash.update(chunk)
                
                md5 = md5_hash.hexdigest()

            try:
                smbfile = session.query(SMBFile) \
                    .filter(SMBFile.filepath==remote_path, SMBFile.md5_sum==md5) \
                    .first()
            except:
                logging.exception("Exception")
                continue

            if (smbfile is None) or (md5 != smbfile.md5_sum):
                logging.info('File {} has changed, updating ...'.format(remote_path))

                with open(f['local_path'], 'rb') as fp:
                    content = fp.read()

                    if content.startswith(codecs.BOM_UTF8):
                        content = content.decode('utf-8-sig')
                    else:
                        content = content.decode('utf-8')

                    smbfile = SMBFile(filepath=remote_path,
                        filesize=int(f['size']),
                        content=content,
                        created=str(f['created']),
                        last_modified=str(f['last_write']),
                        retrieved_at=datetime.datetime.utcnow(),
                        md5_sum=md5)

                    try:
                        session.add(smbfile)
                        session.commit()
                    except:
                        logging.exception("Adding samba file failed")
                        session.rollback()

                # TODO: Check this correctness
                for name in self.posserver_by_name:
                    if not hasattr(servers, name):
                        logging.warning(f'No POS Device {name} found!')
                        continue

                    device = getattr(servers, name)

                    info = self.posserver_by_name[name]
                    if not info.task or info.task.done():
                        logging.warning('Server for {} is stopped, ignore'.format(name))
                        continue

                    match = re.match(device.pattern, filename)
                    if not match:
                        continue

                    logging.info(f"Matched {filename} for {name}")
                    info.set_new_xml(content)
                    await info.notify_xml_changed()


    async def create_servers(self):
        config = Config.get()
        ports_set = set()
        servers = config.servers

        for name in servers.__dict__:
            if name in self.posserver_by_name:
                info = self.posserver_by_name[name]
                if info.task and not info.task.done():
                    continue

            server = getattr(servers, name)
            host = server.host
            port = int(server.port)
            boards = server.boards

            if port in ports_set:
                logging.error('Duplicated port: {}!'.format(port))
                return

            info = PosServer(
                name=name,
                host=host,
                port=port,
                boards=boards)

            self.posserver_by_name[name] = info

        for name in self.posserver_by_name:
            info = self.posserver_by_name[name]
            info.task = asyncio.create_task(info.run())


    async def run_samba(self, name, conf):
        dirpath = tempfile.mkdtemp()
        logging.info(f'Running samba for {name}')

        while True:
            now = datetime.datetime.now()
            year = '{}'.format(now.year % 100)
            month = '{:02d}'.format(now.month)
            day = '{:02d}'.format(now.day)

            template = string.Template(conf.path)
            values = {
                'year': year,
                'month': month,
                'day': day
            }
            path = template.substitute(values)

            logging.info('Seaching files in: {}'.format(path))
            files = await asyncio.get_running_loop().run_in_executor(self.excutor,
                functools.partial(samba_retrieval,
                username=conf.username,
                password=conf.password,
                host=conf.host,
                port=int(conf.port),
                remote_name=conf.remote_name,
                domain=conf.domain,
                service=conf.service,
                path=path,
                localpath=dirpath))

            await self.process_files(files)

            sleep = int(conf.interval)
            logging.info('Sleeping for {} secs'.format(sleep))
            await asyncio.sleep(sleep)

        shutil.rmtree(dirpath)


    async def run(self):
        await self.create_servers()

        tasks = []
        config = Config.get()
        for name in config.samba.__dict__:
            cfg = getattr(config.samba, name)
            if not cfg.enable:
                continue

            tasks.append(asyncio.create_task(self.run_samba(name, cfg)))

        if len(tasks) >0:
            try:
                await asyncio.gather(*tasks)
            except:
                logging.exception("samb's tasks cancelled")
        else:
            await asyncio.sleep(1)


    def main(self):
        try:
            asyncio.run(self.run())
        except KeyboardInterrupt:
            logging.info('Done')

        self.excutor.shutdown(wait=False)


def run_samba():
    Config.init(server_cfg_file())
    setup_logger('samb.txt')

    config = Config.get()
    SessionSingleton.recreate(config.app.database)

    app = App()
    app.main()


if __name__ == '__main__':
    run_samba()
