import sys
import datetime
import logging
import base64
import binascii
import asyncio
import multiprocessing as mp
from dataclasses import dataclass
from concurrent.futures import ProcessPoolExecutor
import functools

from starlette.authentication import (
    AuthenticationBackend, AuthenticationError, SimpleUser, UnauthenticatedUser,
    AuthCredentials, requires
)

from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.responses import Response, JSONResponse, PlainTextResponse, HTMLResponse
from starlette.schemas import SchemaGenerator
from starlette.middleware.authentication import AuthenticationMiddleware

import uvicorn
import dateutil
import dateutil.parser
import json
import aiohttp

from package.server.auth_es import run_auth
from package.shared.logger import setup_logger
from package.shared.config import Config
from package.shared import models
from package.shared.util import \
    web_get_streams, web_search_chunks_by_pos, \
    web_get_states, web_get_gateways, server_cfg_file, \
    get_request_json


class BasicAuthBackend(AuthenticationBackend):
    async def authenticate(self, request):
        if "Authorization" not in request.headers:
            return

        auth = request.headers["Authorization"]
        try:
            scheme, credentials = auth.split()
            if scheme.lower() != 'basic':
                return

            decoded = base64.b64decode(credentials).decode("ascii")
        except (ValueError, UnicodeDecodeError, binascii.Error) as exc:
            raise AuthenticationError('Invalid basic auth credentials')

        username, _, password = decoded.partition(":")
        # TODO: Verify usename/password
        return AuthCredentials(["authenticated"]), SimpleUser(username)


def create_app(debug: bool):
    app = Starlette(debug=debug)
    app.add_middleware(AuthenticationMiddleware, backend=BasicAuthBackend())
    executor: ProcessPoolExecutor = None


    @app.on_event('startup')
    async def startup_handler():
        nonlocal executor

        ctx = mp.get_context('spawn')
        executor = ProcessPoolExecutor(
            max_workers=4,
            mp_context=ctx)


    @app.on_event('shutdown')
    async def shutdown_handler():
        nonlocal executor
        executor.shutdown(wait=False)


    @app.route('/')
    class Homepage(HTTPEndpoint):
        @requires('authenticated')
        async def get(self, request):
            return PlainTextResponse('Welcome from server')

    # Auth
    @app.route('/auth')
    class Auth(HTTPEndpoint):
        async def post(self, request):
            rq = await request.json()
            if 'username' not in rq:
                return Response('Bad request', status_code=400, headers=None, media_type=None)

            if 'password' not in rq:
                return Response('Bad request', status_code=400, headers=None, media_type=None)

            username = rq['username']
            password = rq['password']

            config = Config.get()
            loop = asyncio.get_event_loop()
            kwargs = {
                'host': config.es.host,
                'port': config.es.port,
                'username': username,
                'password': password
            }

            nonlocal executor

            try:
                result = await loop.run_in_executor(
                    executor,
                    functools.partial(run_auth, **kwargs))
            except:
                logging.exception('Run auth exception')

            reply = dict()

            if result is None:
                reply = { 'reason': 'Invalid credentials' }
            else:
                reply = { 'streams': result }

            return JSONResponse(reply)

    
    # List of gateways
    @app.route('/gateways')
    class GatewaysAvailability(HTTPEndpoint):
        async def get(self, request):
            return JSONResponse(web_get_gateways())


    # List of streams
    @app.route('/streams')
    class StreamsAvailability(HTTPEndpoint):
        async def forward(self):
            session = models.SessionSingleton.instance().session
            objs = []

            try:
                gateways = session.query(models.Gateway).all()
            except:
                logging.exception("Exception")
                session.rollback()
            else:
                for gateway in gateways:
                    try:
                        url = 'http://{host}:{port}/streams'.format(
                            host=gateway.address,
                            port=gateway.port)

                        ret = await asyncio.wait_for(get_request_json(url), timeout=5)
                        objs.extend(ret)
                    except asyncio.TimeoutError:
                        logging.warning(f'Gateway {gateway.name} not respond')
                    except Exception as e:
                        logging.warning(e)

            return JSONResponse(objs)

        async def get(self, request):
            try:
                request.query_params['forward']
                return await self.forward()
            except:
                pass

            return JSONResponse(web_get_streams())


    # Stream states
    @app.route('/states')
    class StateEndpoint(HTTPEndpoint):
        async def forward(self):
            session = models.SessionSingleton.instance().session
            obj = dict()

            try:
                gateways = session.query(models.Gateway).all()
            except:
                logging.exception("Exception")
                session.rollback()
            else:
                for gateway in gateways:
                    try:
                        url = 'http://{host}:{port}/states'.format(
                            host=gateway.address,
                            port=gateway.port)

                        ret = await asyncio.wait_for(get_request_json(url), timeout=5)
                        for k, v in ret.items():
                            obj[k] = v

                    except asyncio.TimeoutError:
                        logging.warning(f'Gateway {gateway.name} not respond')
                    except Exception as e:
                        logging.warning(e)

            return JSONResponse(obj)


        async def get(self, request):
            try:
                request.query_params['forward']
                return await self.forward()
            except:
                pass

            return JSONResponse(web_get_states())


    # Return chunks contains value in range (from - to) of a stream (or all streams)
    @app.route('/send_ad')
    class AdEndpoint(HTTPEndpoint):
        async def get(self, request):
            try:
                # Board to send
                board_ip = request.query_params['board_ip']
                board_port = int(request.query_params['board_port'])

                # Pos
                ad_name = request.query_params['ad_name']
                ad_ip = request.query_params['ad_ip']
                ad_port = request.query_params['ad_port']

                # Secret
                secret = request.query_params['secret']
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            if secret != 'AIGSG_13579':
                return Response('Invalid secret', status_code=404, headers=None, media_type=None)

            try:
                async with aiohttp.ClientSession() as session:
                    params = 'cgi/registry.exe?NAME={name}&IP={host}&ETHADDR=None&SHOW=1&PORT={port}&rdm=700622665'.format(
                        name=ad_name,
                        host=ad_ip,
                        port=ad_port)

                    async with session.post(f'http://{board_ip}:{board_port}/{params}', data=None) as response:
                        await response.text()
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            return HTMLResponse('<html><body><strong>Ok</strong></body></html>')


    # Return chunks contains value in range (from - to) of a stream (or all streams)
    @app.route('/search')
    class SearchEndpoint(HTTPEndpoint):
        async def forward(self, min_from, min_to, value):
            session = models.SessionSingleton.instance().session
            obj = []

            try:
                gateways = session.query(models.Gateway).all()
            except:
                logging.exception("Exception")
                session.rollback()
            else:
                for gateway in gateways:
                    if gateway.address is None or len(gateway.address)==0 or gateway.address=='127.0.0.1':
                        continue

                    try:
                        url = 'http://{host}:{port}/search?min_from={min_from}&min_to={min_to}&value={value}'.format(
                            host=gateway.address,
                            port=gateway.port,
                            min_from=min_from,
                            min_to=min_to,
                            value=value)

                        ret = await asyncio.wait_for(get_request_json(url), timeout=5)
                        obj.extend(ret)

                    except asyncio.TimeoutError:
                        logging.warning(f'Gateway {gateway.name} not respond')
                    except Exception as e:
                        logging.warning(e)

            return JSONResponse(obj)


        async def get(self, request):
            try:
                min_from = int(request.query_params['min_from'])
                min_to = int(request.query_params['min_to'])
                value = request.query_params['value']
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            forward = None
            try:
                forward = request.query_params['forward']
            except:
                pass

            if forward:
                return await self.forward(min_from, min_to, value)

            session = models.SessionSingleton.instance().session
            objs = []

            try:
                gateways = session.query(models.Gateway).all()
            except:
                logging.exception("Exception")
                session.rollback()
            else:
                for gateway in gateways:
                    objs.extend(web_search_chunks_by_pos(
                        gateway,
                        min_from,
                        min_to,
                        value))

            return JSONResponse(objs)


    @app.route('/esdata')
    class ESDataEndpoint(HTTPEndpoint):
        async def get(self, request):
            try:
                gateway = request.query_params['gateway']
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            session = models.SessionSingleton.instance().session

            try:
                gw = session.query(models.Gateway).filter(models.Gateway.serial==gateway).first()
            except:
                logging.exception("Exception")
                session.rollback()
                gw = None

            if not gw:
                return Response("Not gateway found", status_code=404, headers=None, media_type=None)

            try:
                pair = session.query(models.Setting).filter(models.Setting.gateway==gw).first()
            except:
                logging.exception("Exception")
                session.rollback()
                pair = None

            if pair is None or len(pair.value)==0:
                return Response("Not setting found", status_code=404, headers=None, media_type=None)

            try:
                txt = pair.value
                obj = json.loads(txt)
            except Exception as e:
                return Response(str(e), status_code=404, headers=None, media_type=None)

            return JSONResponse(obj)

    return app


def run_web():
    Config.init(server_cfg_file())
    setup_logger('web.txt')


    config = Config.get()
    models.SessionSingleton.recreate(config.app.database)

    app = create_app(debug=config.app.debug)
    uvicorn.run(app, host=config.web.host, port=int(config.web.port))


if __name__ == '__main__':
    run_web()